// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_ByRangeDelivery_hpp
#define anna_comm_ByRangeDelivery_hpp

#include <anna/core/functions.hpp>
#include <anna/core/tracing/Logger.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>

#include <anna/comm/RoundRobinDelivery.hpp>

namespace anna {

namespace comm {
class Resource;
}

namespace comm {

/**
   Servicio de reparto de carga permite selecionar el servidor al que enviar la peticion en base al valor de una
   clave previamente indicada.

   A diferencia de los comm::Service habituales la asociacion entre el servicio y el proceso no debe hacerse
   mediante comm::Service::attach, sino mediante comm::ByRangeDelivery::attach.

   \param TKey Tipo de clave usada para indexar los servicios de reparto.
   \param TDelivery: Tipo de reparto usado en cada rango.
   \param TSerializer: Clase que es capaz de convertir la \em TKey en una cadena. Debe implementar un metodo
   \em asString que recibe una instancia de TKey y devuelva una cadena.

   La clase \em Delivery debe implementar los mismos metodos que un comm::Service.
*/
template < typename TKey, typename TDelivery = RoundRobinDelivery, typename TSerializer = anna::functions > class ByRangeDelivery : public comm::Service {
protected:
  struct Range {
    const TKey bottom;
    const TKey top;
    TDelivery* delivery;

    Range(const std::string& baseName, const TKey& _bottom, const TKey& _top) :
      bottom(_bottom), top(_top) {
      std::string name(baseName);
      name += '[';
      name += TSerializer::asString(_bottom);
      name += ',';
      name += TSerializer::asString(_top);
      name += ']';
      delivery = new TDelivery(name.c_str(), false);
    }

    std::string asString() const throw() {
      std::string result("comm::ByRangeDelivery { Bottom: ");
      result += TSerializer::asString(bottom);
      result += " | Top: ";
      result += TSerializer::asString(top);
      result += " | ";
      result += delivery->asString();
      return result += " }";
    }

    xml::Node* asXML(xml::Node* parent) const throw() {
      xml::Node* result = parent->createChild("comm.Range");
      result->createAttribute("Bottom", TSerializer::asString(bottom));
      result->createAttribute("Top", TSerializer::asString(top));
      delivery->asXML(result);
      return result;
    }
  };

public:
  typedef std::vector<Range*> range_container;
  typedef typename range_container::iterator range_iterator;
  typedef typename range_container::const_iterator const_range_iterator;

  /**
     Constructor.
     \param name Nombre logico del servicio de reparto de carga.
     \param isCritical Indica las caracteristicas del reparto de carga. Debe valor \em true si el
     reparto de carga a crear es critico o \em false en otro caso.
  */
  ByRangeDelivery(const char* name, const bool isCritical) :
    comm::Service(name, isCritical),
    a_currentRange(NULL)
  {;}

  /**
     Destructor.
  */
  virtual ~ByRangeDelivery() {
    for(range_iterator ii = range_begin(), maxii = range_end(); ii != maxii; ii ++)
      delete range(ii)->delivery;

    a_ranges.clear();
  }

  /**
     Instancia un rango de reparto con los parametros definidos como parametros. Una vez que
     obtengamos el comm::Delivery asociado al rango creado debemos asociar tantos comm::Resource
     como sea necesario.

     \param bottom Valor minimo del rango.
     \param top Valor maximo del rango.
     \return La instancia del servicio de reparto asociado a este rango.
  */
  range_iterator createRange(const TKey& bottom, const TKey& top)
  throw(RuntimeException) {
    Range* result;

    if(bottom > top) {
      std::string msg(asString());
      msg += " | Bottom: ";
      msg += TSerializer::asString(bottom);
      msg += " | Invalid range (bottom > top)";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    if((result = find(bottom)) != NULL) {
      std::string msg(asString());
      msg += " | Bottom: ";
      msg += TSerializer::asString(bottom);
      msg += " | Overlapped with ";
      msg += result->asString();
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    if((result = find(top)) != NULL) {
      std::string msg(asString());
      msg += " | top: ";
      msg += TSerializer::asString(top);
      msg += " | Overlapped with ";
      msg += result->asString();
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    a_ranges.push_back(result = new Range(getName(), bottom, top));
    return range_begin() + a_ranges.size() - 1;
  }

  /**
     Asocia el comm::Server recibido como parametro al rango representado por el iterador recibido.
     \param ii Iterador de rango al que asociar el servicio. Debe ser resultado de invocar a #createRange.
     \param server Instancia del comm::Server a incorporar dentro del rango.
  */
  void attach(range_iterator& ii, Server* server) throw(RuntimeException) {
    range(ii)->delivery->attach(server);
    comm::Service::attach(server);
  }

  /**
     Establece la clave que se usara para decidir que pareja de maquinas usar para enviar el mensaje.
     \param key Clave usada para calcular el reparto.
     \warning Este metodo debe invocarse siempre antes de usar este servicio de reparto.
  */
  void prepare(const TKey& key) throw(RuntimeException) {
    Range* w = find(key);

    if(w == NULL && Logger::isActive(Logger::Warning)) {
      std::string msg(asString());
      msg += " | TKey: ";
      msg += TSerializer::asString(key);
      msg += " | No range associated";
      Logger::warning(msg, ANNA_FILE_LOCATION);
    } else
      a_currentRange = w;

    if(a_currentRange == NULL) {
      std::string msg(asString());
      msg += " | TKey: ";
      msg += TSerializer::asString(key);
      msg += " | No delivery service associated has been found";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }
  }

  std::string asString() const
  throw() {
    std::string result = className();
    result += " { ";
    result += comm::Delivery::asString();
    result += " | Ranges: ";

    for(const_range_iterator ii = range_begin(), maxii = range_end(); ii != maxii; ii ++)  {
      result += "\n\t";
      result += range(ii)->asString();
    }

    return result += "}";
  }

  xml::Node* asXML(xml::Node* parent) const
  throw() {
    xml::Node* result = parent->createChild("anna.comm.ByRangeDelivery");
    xml::Node* node = comm::Service::asXML(result);

    for(const_range_iterator ii = range_begin(), maxii = range_end(); ii != maxii; ii ++)
      range(ii)->asXML(node);

    return result;
  }

  /**
     Metodo que devuelve el nombre completo de este selector de recursos.
     Para evitar ambigüedades este nombre incluye la lista completa de \em namespaces
     a los que pertenece la clase.
     \return Una cadena con el nombre de este selector.
  */
  static const char* className() throw() { return "anna::comm::ByRangeDelivery"; }

protected:
  range_iterator range_begin() throw() { return a_ranges.begin(); }
  range_iterator range_end() throw() { return a_ranges.end(); }

  const_range_iterator range_begin() const throw() { return a_ranges.begin(); }
  const_range_iterator range_end() const throw() { return a_ranges.end(); }

  static Range* range(range_iterator& ii) throw() { return *ii; }
  static const Range* range(const_range_iterator& ii) throw() { return *ii; }

  virtual void do_initialize()
  throw(RuntimeException) {
    for(range_iterator ii = range_begin(), maxii = range_end(); ii != maxii; ii ++)
      range(ii)->delivery->initialize();
  }

private:
  range_container a_ranges;
  Range* a_currentRange;

  Range* find(const TKey& key) throw() {
    Range* result(NULL);
    Range* w;

    for(range_iterator ii = range_begin(), maxii = range_end(); ii != maxii; ii ++) {
      w = range(ii);

      if(w->bottom <= key && key <= w->top) {
        result = w;
        break;
      }
    }

    return result;
  }

  comm::Resource* do_apply()
  throw(RuntimeException) {
    if(a_currentRange == NULL) {
      std::string msg(asString());
      msg += " | Wrong call to anna::comm::ByRangeDelivery::prepare";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    return a_currentRange->delivery->apply();
  }

  // Considera que el servicio esta NO-disponible cuando TODOS los servicios fallan
  bool do_fault(const comm::Resource* resource)
  throw() {
    int n, nfault;
    n = nfault = 0;

    for(range_iterator ii = range_begin(), maxii = range_end(); ii != maxii; ii ++) {
      n ++;

      if(range(ii)->delivery->fault(resource) == true)
        nfault ++;
    }

    return (n == nfault);
  }

  // Considera que el servicio esta recuperado cuando alguno de los servicios esta disponible
  bool do_recover(const comm::Resource* resource)
  throw() {
    bool result(false);

    for(range_iterator ii = range_begin(), maxii = range_end(); ii != maxii; ii ++) {
      if(range(ii)->delivery->recover(resource) == true)
        result = true;
    }

    return result;
  }
};

}
}

#endif

