// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_CompatCodec_hpp
#define anna_comm_CompatCodec_hpp

#include <map>

#include <anna/core/DataBlock.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/core/util/SortedVector.hpp>

#include <anna/comm/Message.hpp>
#include <anna/comm/Variable.hpp>

namespace anna {

class DataBlock;
class Second;
class Millisecond;
class Microsecond;

namespace comm {

/**
   Codificador/Decodificador compatible con los mensajes

   Esta clase ofrece una forma muy eficaz para codificar estructuras de datos complejas y pretende
   sustituir los tipicos aplanadores y desaplanadores ya que permite que cualquier proceso
   defina facilmente la forma de transferir cierta informacion sobre un bloque de memoria
   y viceversa.

   Vamos a ver un ejemplo de uso. Primero vamos a definir la clase MensajeDemo:

* \code
*
* #include <anna.comm.CompatCodec.h>
*
* class MensajeDemo : public CompatCodec {
* public:
*    static const CompatCodec::Type type = 10;
*
*    MensajeDemo () : CompatCodec (type),
*       a_dataBlock (true)
*    {
*       attach ("Entero", 0, a_entero);
*       attach ("Cadena", 1, a_cadena);
*       attach ("Bloque datos", 2, a_dataBlock);
*    }
*
*    // Accesores
*    const int obtenerEntero () const { return a_entero; }
*    const std::string& obtenerCadena () const { return a_cadena; }
*    const DataBlock& obtenerDataBlock () const { return a_dataBlock; }
*
*    // Modificadores
*    void establecerEntero (const int entero) { a_entero = entero; }
*    void establecerCadena (const std::string& cadena) { a_cadena = cadena; }
*    void establecerCadena (const char* cadena) { a_cadena = cadena; }
*    void establecerDataBlock (const DataBlock& dataBlock) { a_dataBlock = dataBlock; }
*
* private:
*    int a_entero;
*    std::string a_cadena;
*    DataBlock a_dataBlock;
* };
*
* \endcode
*
   La clase que use esta clase para enviar un mensaje debe establecer los valores de cada uno de
   los datos mediante los modificadores definidos y luego invocar al metodo #code que
   transferiria el contenido de las variables asociadas a este mensaje a un bloque de memoria,
   que normalmente seriausado como mensaje.

   Por otra parte la clase que recibe el mensaje invocar�al metodo #decode que transfiere
   el contenido del bloque de memoria a cada una de las variables asociadas al mensaje.
   Posteriormente, podremos acceder al contenido de cada una de las variables asociadas al
   mensaje atraves de los accesores definidos para el caso.

   Estos codificadores nos dan la posibilidad de definir variables opcionales, de forma que
   una determinada variable puede no ser transferida al bloque de memoria, y por tanto puede
   no ser recibida en el otro extremo. Ver el metodo #isNull y #setNull para mas informacion.

   \warning Esta clase no establece proteccion ante accesos concurrentes
*/
class CompatCodec : public Message {
  struct SortById {
    static short int value(const comm::Variable* variable) throw() {
      return variable->getId();
    }
  };

  class VariableContainer {
  public:
    typedef comm::Variable* type_pointer;

    typedef type_pointer* iterator;
    typedef const type_pointer* const_iterator;

    VariableContainer();

    void clear() throw();
    void add(comm::Variable* variable) throw();
    comm::Variable* find(const int id) throw();
    const comm::Variable* find(const int id) const throw();

    int size() const throw() { return a_size; }

    iterator begin() throw() { return a_variables; }
    iterator end() throw() { return a_variables + a_size; }

    const_iterator begin() const throw() { return a_variables; }
    const_iterator end() const throw() { return a_variables + a_size; }

    static comm::Variable* data(iterator ii) throw() { return *ii; }
    static const comm::Variable* data(const_iterator ii) throw() { return *ii; }

  private:
    comm::Variable** a_variables;
    int a_maxSize;
    int a_size;
  };

public:
//   typedef SortedVector <comm::Variable, SortById, short int> container;
  typedef VariableContainer container;
  typedef container::iterator iterator;
  typedef container::const_iterator const_iterator;
  typedef unsigned char Type;

  /**
     Constructor.

     @param type Tipo por el que sera conocido este tipo de mensaje.
     @param scramble Indica si el mensaje debe ser codificado de forma que no se pueda ver el contenido
     del mismo con una simple herramienta de monitorizacion de mensajes de red. Por defecto esta
     activo ya que la codificacion realizada es muy simple y rapida y el tiempo empleado es casi
     inapreciable.
  */
  explicit CompatCodec(const Type type, const bool scramble = true);

  /**
     Destructor.
  */
  virtual ~CompatCodec();

  // Accesores
  /**
     Devuelve el identificador del mensaje indicado en el constructor.
    @return El identificador de este mensaje.
  */
  Type getType() const throw() { return a_type; }

// Metodos
  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos.
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param id Identificador asignado al dato.
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, const short int id, std::string& value) throw(RuntimeException);

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param id Identificador asignado al dato.
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, const short int id, const char*& value) throw(RuntimeException);

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param id Identificador asignado al dato.
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, const short int id, int& value) throw(RuntimeException);

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param id Identificador asignado al dato.
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, const short int id, S64& value) throw(RuntimeException);

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param id Identificador asignado al dato.
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, const short int id, bool& value) throw(RuntimeException);

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param id Identificador asignado al dato.
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa. Debe tener activado el sistema de copia profunda.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, const short int id, DataBlock& value) throw(RuntimeException);

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param id Identificador asignado al dato.
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa. Debe tener activado el sistema de copia profunda.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, const short int id, float& value) throw(RuntimeException);

  /**
     Asocia el valor recibido como parametro al dato interno identificado por `id'. Esta clase
     estapensada principalmente para comunicar entre si procesos remotos
     La unica restriccion que se les imponen a ambos que es compartan el significado que cada uno
     de ellos le da a un determinado `id'.

     \param name Nombre logico de la variable
     @param id Identificador asignado al dato.
     @param value Rerencia a una variable de nuestro entorno usada para transferir los datos de la memoria
     al mensaje y viceversa. Debe tener activado el sistema de copia profunda.
     @return Un puntero que hace referencia al nuevo dato interno que ha sido creado.
  */
  const Variable* attach(const char* name, const short int id, double& value) throw(RuntimeException);

  const Variable* attach(const char* name, const short int id, Second& value) throw(RuntimeException);
  const Variable* attach(const char* name, const short int id, Millisecond& value) throw(RuntimeException);
  const Variable* attach(const char* name, const short int id, Microsecond& value) throw(RuntimeException);

  /**
   * Asocia el mensaje recibido como un parmetro interno especificado por \em id.
   * \param name Nombre lgico de la variable.
   * \param id Identificador del dato asignado.
   * \param value Instancia del mensaje que se codificar/decodificar de forma recursiva.
   */
  const Variable* attach(const char* name, const short int id, comm::CompatCodec& value) throw(RuntimeException);

  /**
     Devuelve la referencia al dato interno identificado por el `id' recibido como parametro.

     @param id Identificador asignado al dato que queremos obtener.

     @return La referencia de la variable identificada por 'id'. Si no existe se lanzar�una excepcin.
  */
  const Variable& find(const short int id) const throw(RuntimeException);

  /**
     Marca el dato asociado al identificador recibido como nulo, lo cual conlleva que el dato no seria
     transferido al bloque de memoria del mensaje en caso de invocar al metodo de codificacion.

     @param id Identificador asignado al dato.
     @param isNull Indica la nueva marca de la variable.
  */
  void setNull(const short int id, const bool isNull = true) throw(RuntimeException);

  /**
     Marca el dato recibido como nulo, lo cual conlleva que el dato no seria transferido al bloque
     de memoria del mensaje en caso de invocar al metodo de codificacion.

     @param variable Instancia de Varaible obtenida al invocar al mtodo #attach
     @param isNull Indica la nueva marca de la variable.
  */
  void setNull(const Variable* variable, const bool isNull = true) throw();

  /**
     @param id Identificador asignado al dato.
  *
     @return Devuelve @em false si el dato identificado por `id' tiene algun valor o @em true en caso de que
     el dato haya sido marcado como nulo. Por defecto se considera que todos los datos tienen valor.
     Un dato puede tener un value nulo, bien por que se ha invocado a la funcin @ref setNull
     o bien porque se recibio un mensaje en el que no venia contenido el identificador del dato.
  */
  bool isNull(const short int id) const throw(RuntimeException);

  /**
     Devuelve un iterador al comienzo de la lista de variables asociados a este mensaje.
     \return Un iterador al comienzo de la lista de variables asociados a este mensaje.
  */
  iterator begin() throw() { return a_variables.begin(); }

  /**
     Devuelve un iterador al comienzo de la lista de variables asociados a este mensaje.
     \return Un iterador al comienzo de la lista de variables asociados a este mensaje.
  */
  const_iterator begin() const throw() { return a_variables.begin(); }

  /**
     Devuelve un iterador al comienzo de la lista de variables asociados a este mensaje.
     \return Un iterador al final de la lista de variables asociados a este mensaje.
  */
  iterator end() throw() { return a_variables.end(); }

  /**
     Devuelve un iterador al comienzo de la lista de variables asociados a este mensaje.
     \return Un iterador al final de la lista de variables asociados a este mensaje.
  */
  const_iterator end() const throw() { return a_variables.end(); }

  /**
     Devuelve el nmero de enginees asociados a.esta instancia.
     \return el nmero de enginees asociados a.esta instancia.
  */
  int size() const throw() { return a_variables.size(); }

  /**
     Devuelve la instancia de la variable sobre el que esta posicionado el iterador recibido
     como parametro.
     \param ii Iterador que debera estar comprendido entre begin y end.
     \return La instancia de la variable sobre el que esta posicionado el iterador recibido
  */
  static Variable* variable(iterator ii) throw() { return container::data(ii); }

  /**
     Devuelve la instancia de la variable sobre el que esta posicionado el iterador recibido
     como parametro.
     \param ii Iterador que debera estar comprendido entre begin y end.
     \return La instancia de la variable sobre el que esta posicionado el iterador recibido
  */
  static const Variable* variable(const_iterator ii) throw() { return container::data(ii); }

  /**
     Transfiene los datos establecidos en este mensaje interno a un bloque de memoria cuya instancia es devuelta por este metodo.
     @return Bloque de memoria que contiene la informacion del mensaje.
  */
  virtual const DataBlock& code() throw(RuntimeException);

  /**
     Transfiene la informacion contenida en el bloque de memoria recibido hacia las variables asociadas a este mensaje interno.
     Las variables cuyo value no esta incluidas en el bloque de memoria seria marcadas como nulas.

     @param dataBlock Bloque de memoria que contiene las variables codificadas.
  */
  virtual void decode(const DataBlock& dataBlock) throw(RuntimeException);

  /**
     Permite conocer el identificador del mensaje que viene contenido en el bloque de memoria
     antes de realizar la decodificacion.

     @param dataBlock Bloque de memoria que contiene la informacion del mensaje.

     @return El tipo del mensaje contenido en el bloque de memoria.
  */
  static Type getType(const DataBlock& dataBlock) throw(RuntimeException);

protected:
  bool a_scramble;

private:
  const Type a_type;
  container a_variables;
  int a_nullCounter;

  static bool st_initScramble;

  CompatCodec(const CompatCodec&);
  CompatCodec& operator = (const CompatCodec&);

  void normalDecode(const char* data, const int size, const int maxdata) throw(RuntimeException);
  bool optimizedDecode(const char* data, const int size) throw(RuntimeException);
};

}
}

#endif
