// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_Network_hpp
#define anna_comm_Network_hpp

#include <netinet/in.h>

#include <vector>

#include <anna/core/Singleton.hpp>

#include <anna/comm/INetAddress.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace comm {

class Host;
class Device;
class Server;
class TransportFactory;
class ReceiverFactory;

/**
   Representacion logica de la estructura de red donde se ejecuta nuestra aplicacion.
*/
class Network : public Singleton <Network> {
public:
  /**
   * Modo de actuar a la hora de crear una conexin mediante #createConnection o #resolveConnection.
   * \li Si el modo es \em Unique y ya existe una instancia previa conectada a una IP puerto se devuelve esa misma
   * instancia,
   * \li Si el modo es \em Multiple se pueden abrir tantas conexiones como se dese sobre una misma IP:port.
   */
  struct Port { enum _v { Unique, Multiple }; };
  struct DoConnect { enum _v { Yes, No }; };


  typedef std::vector <Host*> host_container; /**< Definicion para gestionar las maquinas */
  typedef host_container::iterator host_iterator; /**< Definicion para el iterador de maquinas */
  typedef host_container::const_iterator const_host_iterator; /**< Definicion para el iterador de maquinas */

  typedef std::vector <Device*> device_container; /**< Definicion para gestionar los dispositivos de red */
  typedef device_container::iterator device_iterator; /**< Definicion para el iterador de dispositivos de red */
  typedef device_container::const_iterator const_device_iterator; /**< Definicion para el iterador de dispositivos de red */

  /**
     Devuelve un puntero al dispositivo que coincide con la direccion IP
     recibida como parametro. Si no encuentra ninguna coincidencia se creara automaticamente.

     @param address Direccion de la maquina buscada.

     @return La instancia del dispositivo que coincide con la direccion IP recibida como parametro.
  */
  Device* find(const in_addr_t& address) throw();

  /**
     Devuelve un iterador al comienzo de la lista de dispositivos de red.
     \return un iterador al comienzo de la lista de dispositivos de red.
  */
  const_device_iterator device_begin() const throw() { return a_devices.begin(); }

  /**
     Devuelve un iterador al final de la lista de dispositivos de red.
     \return un iterador al final de la lista de dispositivos de red.
  */
  const_device_iterator device_end() const throw() { return a_devices.end(); }

  /**
     Devuelve un puntero al elemento sobre el que se encuentra el iterador pasado como
     parametro.
     \param ii Iterador que estamos recorriendo.
     \return un puntero al elemento sobre el que se encuentra el iterador pasado como
     parametro.
  */
  static const Device* device(const_device_iterator ii) throw() { return *ii; }

  /**
     Devuelve un iterador al comienzo de la lista de dispositivos de red.
     \return un iterador al comienzo de la lista de dispositivos de red.
  */
  device_iterator device_begin() throw() { return a_devices.begin(); }

  /**
     Devuelve un iterador al final de la lista de dispositivos de red.
     \return un iterador al final de la lista de dispositivos de red.
  */
  device_iterator device_end() throw() { return a_devices.end(); }

  /**
     Devuelve un puntero al elemento sobre el que se encuentra el iterador pasado como
     parametro.
     \param ii Iterador que estamos recorriendo.
     \return un puntero al elemento sobre el que se encuentra el iterador pasado como
     parametro.
  */
  static Device* device(device_iterator ii) throw() { return *ii; }

  /**
     Realiza una busqueda secuencial entre todas las maquinas y devuelve la instancia de la
     maquina asociada al nombre recibido como parametro. Si no existia una instancia registrada
     con este nombre se creara.

     @param name Nombre logico de la maquina.

     @return La instancia de la maquina asociada al nombre recibido.
  */
  Host* find_host(const char* name) throw();

  /**
     Realiza una busqueda secuencial entre todas las maquinas y devuelve la instancia de la
     maquina asociada al nombre recibido como parametro. Si no existia una instancia registrada
     con este nombre se creara.

     @param name Nombre logico de la maquina.

     @return La instancia de la maquina asociada al nombre recibido.
  */
  Host* find_host(const std::string& name) throw() { return find_host(name.c_str()); }

  /**
   * Resuelve el nombre de la maquina recibido como parametro y devuelve la instancia
   * del Host asociado a ese nombre. Si el nombre de host ho ha sido definido previamente mediante
   * el uso de los metodos #find devolvera una instancia de Host que tiene asignada todas las
   * direcciones IP's retornadas por el sistema.
   *
   * \param hostname Nombre logico del servidor que sera usado para resolver. Podria ser una cadena
   * de la forma www.gopher.net
   *
   * \return Si el nombre de host ho ha sido definido previamente mediante
   * el uso de los metodos #find_host devolvera una instancia de Host que tiene asignada todas las
   * direcciones IP's retornadas por el sistema
   *
   * \see man gethostbyname.
   */
  Host* resolve(const char* hostname) throw(RuntimeException);

  /**
   * Resuelve el nombre de la maquina recibido como parametro y devuelve la instancia
   * del Host asociado a ese nombre. Si el nombre de host ho ha sido definido previamente mediante
   * el uso de los metodos #find devolvera una instancia de Host que tiene asignada todas las
   * direcciones IP's retornadas por el sistema.
   *
   * \param hostname Nombre logico del servidor que sera usado para resolver. Podria ser una cadena
   * de la forma www.gopher.net
   *
   * \return Si el nombre de host ho ha sido definido previamente mediante
   * el uso de los metodos #find_host devolvera una instancia de Host que tiene asignada todas las
   * direcciones IP's retornadas por el sistema
   *
   * \see man gethostbyname.
   */
  Host* resolve(const std::string& hostname) throw(RuntimeException) { return resolve(hostname.c_str()); }

  /**
     Devuelve un iterador al comienzo de la lista de maquinas no modificables.
     \return Un iterador al comienzo de la lista de maquinas no modificables.
  */
  const_host_iterator host_begin() const throw() { return a_hosts.begin(); }

  /**
     Devuelve un iterador al final de la lista de maquinas no modificables.
     \return Un iterador al final de la lista de maquinas no modificables.
  */
  const_host_iterator host_end() const throw() { return a_hosts.end(); }

  /**
     Devuelve un puntero al elemento sobre el que se encuentra el iterador pasado como
     parametro.
     \param ii Iterador que estamos recorriendo.
     \return un puntero al elemento sobre el que se encuentra el iterador pasado como
     parametro.
  */
  static const Host* host(const_host_iterator ii) throw() { return *ii; }

  /**
     Devuelve un iterador al comienzo de la lista de maquinas no modificables.
     \return Un iterador al comienzo de la lista de maquinas no modificables.
  */
  host_iterator host_begin() throw() { return a_hosts.begin(); }

  /**
     Devuelve un iterador al final de la lista de maquinas no modificables.
     \return Un iterador al final de la lista de maquinas no modificables.
  */
  host_iterator host_end() throw() { return a_hosts.end(); }

  /**
     Devuelve un puntero al elemento sobre el que se encuentra el iterador pasado como
     parametro.
     \param ii Iterador que estamos recorriendo.
     \return un puntero al elemento sobre el que se encuentra el iterador pasado como
     parametro.
  */
  static Host* host(host_iterator ii) throw() { return *ii; }

  /**
     Crea la instancia de un anna::comm::Server disponible para conectar con la
     IP y puerto indicados.

     \param ip Direccion IP en la que escucha el proceso con el que queremos conectar.
     \param remotePort Puerto remoto en el que atiendo peticiones el proceso con el que conectar.
     \param autoRecovery Indica si en caso de caida se debe intentar la recuperacion
     automatica de la conexion.
     \param transportFactory Factoria de protocolos de transporte usada por los ClientSocket asociados a este
     proceso servidor.
     \param mode Modo de actuar en caso de que ya haya definida una conexin previa contra una misma IP:port
     \param doConnect Realiza o ignora, la conexion del recurso creado.
     \return La instancia de comm::Server asociado al IP y puerto recibido.
     \warning Con modo de puerto unico, si ya existe un proceso definido sobre esa misma IP:port retorna la misma instancia.
  */
  Server* createServer(const char* ip, const int remotePort, const bool autoRecovery, TransportFactory* transportFactory = NULL, const Port::_v mode = Port::Multiple, const DoConnect::_v doConnect = DoConnect::Yes)
  throw(RuntimeException);

  /**
     Crea la instancia de un anna::comm::Server disponible para conectar con la
     IP y puerto indicados.

     \param ip Direccion IP en la que escucha el proceso con el que queremos conectar.
     \param remotePort Puerto remoto en el que atiendo peticiones el proceso con el que conectar.
     \param autoRecovery Indica si en caso de caida se debe intentar la recuperacion
     automatica de la conexion.
     \param receiverFactory Factoria de receptores usada por el comm::ClientSocket usado por el comm::Server a crear.
     \param transportFactory Factoria de protocolos de transporte usada por los ClientSocket asociados a este
     proceso servidor.
     \param mode Modo de actuar en caso de que ya haya definida una conexin previa contra una misma IP:port
     \param doConnect Realiza o ignora, la conexion del recurso creado.
     \return La instancia de comm::Server asociado al IP y puerto recibido.
     \warning Con modo de puerto unico, si ya existe un proceso definido sobre esa misma IP:port retorna la misma instancia.
  */
  Server* createServer(const char* ip, const int remotePort, const bool autoRecovery, ReceiverFactory& receiverFactory, TransportFactory* transportFactory = NULL, const Port::_v mode = Port::Multiple, const DoConnect::_v doConnect = DoConnect::Yes)
  throw(RuntimeException);

//   /**
//      Devuelve la instancia del anna::comm::Server asociado a la IP y puerto recibidos.
//
//      \param ip Direccion IP en la que escucha el proceso con el que queremos conectar.
//      \param remotePort Puerto remoto en el que atiendo peticiones el proceso con el que conectar.
//
//      \return La instancia de comm::Server asociado al IP y puerto recibido.
//      \warning El anna::comm::Server devuelto puede ser NULL.
//   */
//   Server* findServer (const char* ip, const int remotePort) throw (RuntimeException);

  /**
     Crea la instancia de un anna::comm::Server disponible para conectar con la
     IP y puerto indicados.

   * \param hostname Nombre logico del servidor que sera usado para resolver. Podria ser una cadena
   * de la forma www.gopher.net
     \param remotePort Puerto remoto en el que atiendo peticiones el proceso con el que conectar.
     \param autoRecovery Indica si en caso de caida se debe intentar la recuperacion
     automatica de la conexion.
     \param transportFactory Factoria de protocolos de transporte usada por los ClientSocket asociados a este
     proceso servidor.
     \param mode Modo de actuar en caso de que ya haya definida una conexin previa contra una misma IP:port
     \param doConnect Realiza o ignora, la conexion del recurso creado.
     \return La instancia de comm::Server asociado al IP y puerto recibido.
     \warning Con modo de puerto unico, si ya existe un proceso definido sobre esa misma IP:port retorna la misma instancia.
  */
  Server* resolveServer(const char* hostname, const int remotePort, const bool autoRecovery, TransportFactory* transportFactory = NULL, const Port::_v mode = Port::Multiple, const DoConnect::_v doConnect = DoConnect::Yes)
  throw(RuntimeException);

  /**
     Crea la instancia de un anna::comm::Server disponible para conectar con la
     IP y puerto indicados.

   * \param hostname Nombre logico del servidor que sera usado para resolver. Podria ser una cadena
   * de la forma www.gopher.net
     \param remotePort Puerto remoto en el que atiendo peticiones el proceso con el que conectar.
     \param autoRecovery Indica si en caso de caida se debe intentar la recuperacion
     automatica de la conexion.
     \param receiverFactory Factoria de receptores usada por el comm::ClientSocket usado por el comm::Server a crear.
     \param transportFactory Factoria de protocolos de transporte usada por los ClientSocket asociados a este
     proceso servidor.
     \param mode Modo de actuar en caso de que ya haya definida una conexin previa contra una misma IP:port
     \param doConnect Realiza o ignora, la conexion del recurso creado.
     \return La instancia de comm::Server asociado al IP y puerto recibido.
     \warning Con modo de puerto unico, si ya existe un proceso definido sobre esa misma IP:port retorna la misma instancia.
  */
  Server* resolveServer(const char* hostname, const int remotePort, const bool autoRecovery, ReceiverFactory& receiverFactory, TransportFactory* transportFactory = NULL, const Port::_v mode = Port::Multiple, const DoConnect::_v doConnect = DoConnect::Yes)
  throw(RuntimeException);


  /**
   * Obtiene la INetAddress correspondiente a la IP y puerto recibidos como parmetro.
   * \param ip Direccin IP en formato a.b.c.d
   * \param port Puerto de la direccin de red.
   * \return la INetAddress correspondiente a la IP y puerto recibidos como parmetro.
   */
  INetAddress getINetAddress(const char* ip, const int port) throw(RuntimeException);

  /**
   * Obtiene la INetAddress correspondiente a la IP y puerto recibidos como parmetro.
   * \param ip Direccin IP en formato a.b.c.d
   * \param port Puerto de la direccin de red.
   * \return la INetAddress correspondiente a la IP y puerto recibidos como parmetro.
   */
  INetAddress getINetAddress(const std::string& ip, const int port) throw(RuntimeException);


  /**
     Devuelve una cadena con la informacin referente a esta instancia.
     \param parent Nodo XML del que dependende la informacion.
     @return Una cadena con la informacin referente a esta instancia.
  */
  xml::Node* asXML(xml::Node* parent) const throw();

private:
  host_container a_hosts;
  device_container a_devices;
  Host* a_cacheHost;
  Device* a_cacheDevice;

  Network() : a_cacheHost(NULL), a_cacheDevice(NULL) {;}
  Network(const Network&);

  friend class Singleton<Network>;
};

}
}

#endif

