// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_Cloneable_hpp
#define anna_core_Cloneable_hpp

#include <typeinfo>

#include <anna/core/functions.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/core/AutoPointer.hpp>

namespace anna {

/**
 * Clase de la que deben heredar todas las clases con capacidades de clonado.
 *
 * La clase que implemente este interface debe invocar a alguno de las macros que facilitan la implementacin de los mtodos virtuales requeridos.
 *
 * \see \ref clone_final \ref clone_default \ref clone_abstract
 */
class Cloneable : public AutoPointer {
public:
  /**
   * Destructor.
   */
  virtual ~Cloneable() {;}

  /**
   * Genera de esta instancia. Por cada clon generado habr que invocar a #release.
   * \return Un clon de la instancia recibida
   */
  virtual Cloneable* clone() const throw() = 0;

  /**
   * Devuelve el nombre lgico de esta clase.
   * \return el nombre lgico de esta clase.
   */
  static const char* className() throw() { return "Cloneable"; }

protected:
  /**
   * Constructor
   * \param whenFinished Indica que operacin a realizar cuando se invoque a anna::Cloneable::release.
   */
  Cloneable(const WhenFinished::_v whenFinished = WhenFinished::Ignore) : AutoPointer(whenFinished) {;}

  /**
   * Constructor copia.
   * \param other Instancia de la que copiar.
   * \warning la instancia generada con este constructor establecer que al invocar a anna::Cloneable::release se
   * invocar al operador \em delete de esta instancia.
   */
  Cloneable(const Cloneable& other) : AutoPointer(WhenFinished::Delete) {;}
};

/**
 * \page clone_final
 * Define el mtodo \em clone que invoca al constructor copia de la forma:
 *
 * \code
#define anna_clone_final(Class) \
   Class* clone () const throw () { return new Class (*this); } \
   friend class AutoPointer;
 * \endcode
 */
#define anna_clone_final(Class) \
   Class* clone () const throw () { return new Class (*this); } \
   friend class AutoPointer;

/**
 * \page clone_default
 * Define el mtodo \em clone, pero permite que sea re-escrito por herencias posteriores con la forma:
 *
 * \code
#define anna_clone_default(Class) \
   virtual Class* clone () const throw () { return new Class (*this); } \
   friend class AutoPointer;
 * \endcode
 */
#define anna_clone_default(Class) \
   virtual Class* clone () const throw () { return new Class (*this); } \
   friend class AutoPointer;

/**
 * \page clone_abstract
 * Define el mtodo \em clone que tiene que ser re-escrito por herencias posteriores, pero facilita la conversin de
 * la instancia creada al tipo adecuado de 'Cloneable', tendr la forma:
 * \code
#define anna_clone_abstract(Class) \
   virtual Class* clone () const throw () = 0; \
   friend class AutoPointer;
 * \endcode
 */
#define anna_clone_abstract(Class) \
   virtual Class* clone () const throw () = 0; \
   friend class AutoPointer;

}

#endif


