// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_http_wims20_Abstract_hpp
#define anna_http_wims20_Abstract_hpp

#include <vector>

#include <anna/core/RuntimeException.hpp>
#include <anna/core/util/Recycler.hpp>

namespace anna {

namespace http {

namespace wims20 {

/**
   Permite interpretar una URI segn las recomendaciones de WIMS 2.0, lo que facilita
   el desarrollo de aplicaciones integradas en Web 2.0; estas recomendaciones indican
   cmo debe formarse la peticin Abstract (REpresentational State Transfer) para permitir
   el desarrollo de cualquier servicio.

   El formato general de una URI segn la recomendacin de WIMS 2.0 es:

   <p>
http://domain-openapis/path-openapis/serviceID/guid/other_possible_levels?query_parameters
   </p>

   Dnde los campos tienen siguen la siguiente especificacin:
   \li http://domain-openapis: Identifica el recurso del Open API. Formar parte de la configuracin
   de nuestro API (servicio) particular.
   \li path-openapis: Recurso opcional que ajusta la ruta hacia los recursos de ste API. Formar parte
   de la configuracin de nuestro API (servicio) particular.
   \li serviceID: Identificador de recurso.
   \li guid: Identificador del usuario que solicita la peticin.
   \li other_possible_level: Opcionalmente se pueden indicar tantos niveles jerrquicos como fuera
   necesario para el servicio.
   \li query_parameters: Lista de parmetros. Si hay ms de un parmetro se separar con '&'.
*/
class Abstract {
public:
  typedef std::vector <std::string*> other_level_container;
  typedef other_level_container::iterator other_level_iterator;
  typedef other_level_container::const_iterator const_other_level_iterator;

  /**
   * Los parmetros se ordenan en el mismo orden en que fueron indicados
   * por eso no se guardan sobre un std::map, ya que al volcarlos sobre la
   * cadena que actuar como URI apareceran ordenados alfabticamente y
   * quizs el servidor no lo espera as.
   */
  typedef std::pair <std::string*, std::string*> parameter_pkv;
  typedef std::vector <parameter_pkv> parameter_container;
  typedef parameter_container::iterator parameter_iterator;
  typedef parameter_container::const_iterator const_parameter_iterator;

  /**
   * Destructor
   */
  virtual ~Abstract();

  /**
   * Devuelve el campo \em domain-openapis establecido en el contructor.
   * \return el campo \em domain-openapis establecido en el contructor.
   */
  const std::string& getDomain() const throw() { return a_domain; }

  /**
   * Devuelve el campo \em path-openapis
   * \return El campo \em path-openapis, puede ser NULL.
   */
  const std::string* getPath() const throw() { return a_path; }

  /**
   * Devuelve el servicio de la OpenAPI.
   * \param Identificador de servicio usado en la OpenAPI.
   */
  const std::string& getServiceID() const throw() { return a_serviceID; }

  /**
   * Devuelve identificador de usuario que interacciona con el servicio.
   * \return El identificador de usuario que interacciona con el servicio.
   */
  const std::string& getGUID() const throw() { return a_guid; }

  /**
   * Establece el servicio de la OpenAPI.
   * \param serviceID Identificador de servicio usado en la OpenAPI.
   */
  void setServiceID(const std::string& serviceID) throw() { a_serviceID = serviceID; a_fixedPart.clear(); }

  /**
   * Establece el identificador de usuario que interacciona con el servicio.
   * \param guid Identificador de usuario. Dnde por usuario se entiende cualquier elemento
   * que pueda intereccionar con nuestro servicio
   */
  void setGUID(const std::string& guid) throw() { a_guid = guid; a_fixedPart.clear(); }

  /**
   * Devuelve \em true si la estructura contiene parmetros o \em false en otro caso.
   * \return \em true si la estructura contiene parmetros o \em false en otro caso.
   */
  bool hasParameters() const throw() { return a_parameters != NULL && a_parameters->empty() == false; }

  /**
   * Devuelve \em true si la estructura contiene niveles opcionales o \em false en otro caso.
   * \return \em true si la estructura contiene niveles opcionales o \em false en otro caso.
   */
  bool hasOtherLevels() const throw() { return a_otherLevels != NULL && a_otherLevels->empty() == false; }

  /**
   * Limpia el contenido asociado al parmetro \em other_possible_level. Slo debera
   * invocarse a este mtodo en caso de que el servicio destino de la peticin haya cambiado.
   */
  virtual void clearOtherLevels() throw();

  /**
   * Limpia el contenido asociado a los parmetros. Slo debera invocarse a ste mtodo en caso
   * de que el nmero de parmetros a enviar sea distinto al de la peticin anterior.
   * Si son los mismos parmetros con el mismo nombre, deberamos reutilizar el mximo nmero
   * de veces.
   */
  virtual void clearParameters() throw();

  /**
   * Inicializa los toda la informacin asociada a esta instancia.
   */
  void clear() throw() {
    clearOtherLevels();
    clearParameters();
  }

  /**
   * Devuelce una cadena con la informacin relevante sobre esta clase.
   * \return una cadena con la informacin relevante sobre esta clase.
   */
  std::string asString() const throw();

  /**
   * Devuelve un iterator al comienzo de la lista de niveles adicionales.
   * \return un iterator al comienzo de la lista de niveles adicionales.
   * \warning Slo se puede invocar a este mtodo si #hasOtherLevels devuelve \em true.
   */
  other_level_iterator other_level_begin() throw() { return a_otherLevels->begin(); }

  /**
   * Devuelve un iterator al final de la lista de niveles adicionales.
   * \return un iterator al final de la lista de niveles adicionales.
   * \warning Slo se puede invocar a este mtodo si #hasOtherLevels devuelve \em true.
   */
  other_level_iterator other_level_end() throw() { return a_otherLevels->end(); }

  /**
   * Devuelve el valor asociado al iterador.
   * \param ii Iterador sobre los niveles opcionales.
   * \return el valor asociado al iterador.
   */
  static std::string* otherLevel(other_level_iterator ii) throw() { return *ii; }

  /**
   * Devuelve un iterator al comienzo de la lista de niveles adicionales.
   * \return un iterator al comienzo de la lista de niveles adicionales.
   * \warning Slo se puede invocar a este mtodo si #hasOtherLevels devuelve \em true.
   */
  const_other_level_iterator other_level_begin() const throw() { return a_otherLevels->begin(); }

  /**
   * Devuelve un iterator al final de la lista de niveles adicionales.
   * \return un iterator al final de la lista de niveles adicionales.
   * \warning Slo se puede invocar a este mtodo si #hasOtherLevels devuelve \em true.
   */
  const_other_level_iterator other_level_end() const throw() { return a_otherLevels->end(); }

  /**
   * Devuelve el valor asociado al iterador.
   * \param ii Iterador sobre los niveles opcionales.
   * \return el valor asociado al iterador.
   */
  static const std::string& otherLevel(const_other_level_iterator ii) throw() { return **ii; }

  /**
   * Devuelve un iterator al comienzo de la lista de niveles adicionales.
   * \return un iterator al comienzo de la lista de niveles adicionales.
   * \warning Slo se puede invocar a este mtodo si #hasParameters devuelve \em true.
   */
  const_parameter_iterator parameter_begin() const throw() { return a_parameters->begin(); }

  /**
   * Devuelve un iterator al final de la lista de niveles adicionales.
   * \return un iterator al final de la lista de niveles adicionales.
   * \warning Slo se puede invocar a este mtodo si #hasParameters devuelve \em true.
   */
  const_parameter_iterator parameter_end() const throw() { return a_parameters->end(); }

  /**
   * Devuelve el nombre del parmetro asociado al iterador.
   * \param ii Iterador sobre los niveles opcionales.
   * \return el nombre del parmetro asociado al iterador.
   */
  static const std::string& parameter_name(const_parameter_iterator ii) throw() { return *(ii->first); }

  /**
   * Devuelve el valor del parmetro asociado al iterador.
   * \param ii Iterador sobre los niveles opcionales.
   * \return el valor del parmetro asociado al iterador.
   */
  static const std::string& parameter_value(const_parameter_iterator ii) throw() { return *(ii->second); }

protected:
  other_level_container* a_otherLevels;
  parameter_container* a_parameters;

  /**
   * Contructor indicando el parmetro opcional \em path-openapis. Estos dos parmetros se obtendr como
   * parte de la configuracin de nuestro sistema.
   * \param domain: Identifica el recurso del OpenAPI.
   * \param path: Parmetro opcional que ajusta la ruta hacia los recusos de ste API.
   */
  Abstract(const char* whatis, const std::string& domain, const std::string& path);

  /**
   * Constructor que no usar el parmetro opcional \em path-openapis. Este parmetro se obtendr como
   * parte de la configuracin de nuestro sistema.
   * \param domain: Identifica el recurso del OpenAPI.
   */
  explicit Abstract(const char* whatis, const std::string& domain);

  /**
   * Calcula la parte fija de la peticin en base a #calculeShortFixedPart, el \em serviceID y el \em GUID.
   * Mientras estos dos ltimos campos se mantengan constrantes, el resultado de este mtodo no cambia.
   * \return Una cadena con la parte fija de la peticin.
   */
  const std::string& calculeFixedPart() throw(RuntimeException);

  /**
   * Calcula la parte fija corta de la peticin. Tiene en cuenta el \em domain-openapis y si existe
   * el path-openapis. Si fuera necesario incluye el identificador de protocolo "http://".
   * \return Una cadena con la parte fija corta de la peticin.
   */
  const std::string& calculeShortFixedPart() throw(RuntimeException);

  /**
   * Optimiza la creacin y liberacin de cadenas que usa el proceso de interpretacin continuamente.
   * \warning Exclusivamente uso interno.
   */
  std::string* createString() throw(RuntimeException) { return a_string_pool.create(); }

  /**
   * Optimiza la creacin y liberacin de cadenas que usa el proceso de interpretacin continuamente.
   * \warning Exclusivamente uso interno.
   */
  std::string* createString(const char* value) throw(RuntimeException) {
    std::string* result = a_string_pool.create();
    *result = value;
    return result;
  }

  /**
   * Optimiza la creacin y liberacin de cadenas que usa el proceso de interpretacin continuamente.
   * \warning Exclusivamente uso interno.
   */
  std::string* createString(const std::string& value) throw(RuntimeException) { return createString(value.c_str()); }

  /**
   * Optimiza la creacin y liberacin de cadenas que usa el proceso de interpretacin continuamente.
   * \warning Exclusivamente uso interno.
   */
  void destroyString(std::string* str) throw() { a_string_pool.release(str); }

  /**
   * Ampla la lista de niveles.
   * \param otherLevel Nombre del nivel con el que ampliar las lista.
   */
  void other_level_add(const std::string& otherLevel) throw(RuntimeException);

  /**
   * Ampla la lista de parmetros con una nueva pareja (Nombre, Valor).
   * \param name Nombre del parmetro a crear.
   * \param value Valor asociado al parmetro.
   */
  void parameter_set(const std::string& name, const std::string& value) throw(RuntimeException);

  /**
   * Devuelve un iterator al comienzo de la lista de niveles adicionales.
   * \return un iterator al comienzo de la lista de niveles adicionales.
   * \warning Slo se puede invocar a este mtodo si #hasParameters devuelve \em true.
   */
  parameter_iterator parameter_begin() throw() { return a_parameters->begin(); }

  /**
   * Devuelve un iterator al final de la lista de niveles adicionales.
   * \return un iterator al final de la lista de niveles adicionales.
   * \warning Slo se puede invocar a este mtodo si #hasParameters devuelve \em true.
   */
  parameter_iterator parameter_end() throw() { return a_parameters->end(); }

  /**
   * Devuelve el nombre del parmetro asociado al iterador.
   * \param ii Iterador sobre los niveles opcionales.
   * \return el nombre del parmetro asociado al iterador.
   */
  static std::string* parameter_name(parameter_iterator ii) throw() { return ii->first; }

  /**
   * Devuelve el valor del parmetro asociado al iterador.
   * \param ii Iterador sobre los niveles opcionales.
   * \return el valor del parmetro asociado al iterador.
   */
  static std::string* parameter_value(parameter_iterator ii) throw() { return ii->second; }

  /**
   * Concatena las cadenas recibidas teniendo en entre ambas debe de haber un carcter '/'.
   */
  static void appendWithSlash(std::string& target, const std::string& other) throw();

private:
  const std::string a_whatis;
  const std::string a_domain;
  std::string* a_path;
  std::string a_serviceID;
  std::string a_guid;
  Recycler <std::string> a_string_pool;
  std::string a_fixedPart;
};


}
}
}

#endif
