// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_statistics_Engine_hpp
#define anna_statistics_Engine_hpp

#include <anna/core/Singleton.hpp>
#include <anna/core/util/Millisecond.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/core/mt/Mutex.hpp>

// Standard
#include <string>
#include <map>

// Local
#include <anna/statistics/Accumulator.hpp>

namespace anna {
namespace xml {
class Node;
}
}



namespace anna {

namespace statistics {


class Accumulator;


typedef struct {

  std::string SampleFile; // Sample file to optional writtings
  std::string Description;
  std::string Unit;
  bool IntegerNatureSample;

} _concept_identification_t;

typedef std::map <int, _concept_identification_t> _concept_identification_map_t;
typedef std::map <int, _concept_identification_t>::const_iterator _concept_identification_map_iter;
typedef std::map <int, _concept_identification_t>::iterator _concept_identification_map_nc_iter;

typedef std::map <std::string, Accumulator*> _accumulator_map_t;
typedef std::map <std::string, Accumulator*>::const_iterator _accumulator_map_it;
typedef std::map <std::string, Accumulator*>::iterator _accumulator_map_nc_it;


/**
* Class used to configure general behaviour of statistics processing
* By default, the engine is disabled and #enable must be invoked
*/

class Engine : public anna::Singleton <Engine> {


public:

  /** Destructor */
  ~Engine();

  // Sets

  /**
  * Adds a new statistic concept (for example "Time_between_processA_and_processB", "Database time", etc)
  *
  * @param description Concept description
  * @param unit Concept unit description
  * @param integerNatureSample Most of cases we will measure 'time' with the unit which force integer values
  *                            (is more intuitive 850 msecs than 0,850 secs). Then, it is @em true by default.
  *                            This is useful to advice better representation for some indicators like minimum/maximum
  *                             within integer samples.
  *
  * @return Assigned concept identification number (sequence)
  */
  int addConcept(const std::string & description, const std::string & unit, const bool & integerNatureSample = true) throw();


  /**
  * Stops statistics engine
  */
  void enable(void) throw() { a_enabled = true; }


  /**
  * Starts statistics engine
  */
  void disable(void) throw() { a_enabled = false; }

  /**
  * Enable sample log for statistics processings. Engine starts with this feature disabled. When a new concept id is added
  * to the engine, sample log is also disabled for it.
  *
  * Sample logs are used to dump each processing for a certain concept identification (<unix timestamp>,<processed value>).
  *
  *
  * @param id Concept identification. If -1 value is provided, all concepts will be activated.
  * @param sampleFileName Absolute or relative to execution path, and completed with extension '.<concept id>.csv'. Empty string, disables log. Default file name is 'sample'
  *
  * @return @em false if not concept is registered with provided id
  * @warning Many systems add concepts dynamically. This method only affects to current concepts registered at statistics engine.
  */
  bool enableSampleLog(const int & id = -1, const char *sampleFileName = NULL) throw();


  /**
  * Disable sample log for statistics processings
  *
  * @param id Concept identification. If -1 value is provided, all concepts will be deactivated.
  *
  * @return @em false if not concept is registered with provided id
  * @warning Many systems add concepts dynamically. This method only affects to current concepts registered at statistics engine.
  */
  bool disableSampleLog(const int & id = -1) throw();

  // Gets

  /**
  * Gets statistic concept information.
  *
  * @param id Concept identification.
  * @param description Concept description returned by reference
  * @param unit Concept unit description returned by reference
  * @param integerNatureSample nature returned by reference
  *
  * @return @em false if not concept is registered with provided id
  */
  bool getConcept(const int & id, std::string & description, std::string & unit, bool & integerNatureSample) const throw();


  /**
  * Boolean about engine state (enabled / disabled)
  */
  bool enabled(void) const throw() { return (a_enabled); }


  /**
   * Creates a statistic accumulator
   *
   * @param name Accumulator name
   *
   * \return The accumulator created or exception when already exists for the provided name.
   */
  Accumulator *createAccumulator(const std::string &name) throw(anna::RuntimeException);

  /**
   * Returns accumulator instance identified by name.
   *
   * \param name Accumulator name
   *
   * \return The accumulator instance identified by name provided, NULL if not found
   */
  Accumulator* getAccumulator(const std::string &name) throw();



  /**
  * Class string representation
  *
  * @return String with class content
  */
  std::string asString(void) const throw();


  /**
  * Class XML representation
  *
  * @param numberOfDecimals Number of float decimals at XML representation. Default is 2.
  *
  * @return XML with class content
  */
  anna::xml::Node* asXML(anna::xml::Node* parent, const int & numberOfDecimals = 2) const throw();


private:

  Engine(); // private constructor

  _concept_identification_map_t a_concept_identification_map;
  _accumulator_map_t a_accumulators; // you could create accumulators regardless the engine, but this is easier and asXML will show all the information easily
  bool a_enabled;
  int a_sequence_concept_id;
  anna::Mutex a_mutex; // for logSample

  bool logSample(const int & conceptId, const anna::Millisecond & unixTimestamp, const double & value) const throw();

  friend class anna::Singleton <Engine>;
  friend class Accumulator;
};


}
}

#endif
