// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Local
#include <anna/diameter/codec/Message.hpp>
#include <anna/diameter/codec/Format.hpp>

#include <anna/diameter/codec/functions.hpp> // REQUIRED_WORDS
#include <anna/config/defines.hpp> // general types, decoding helpers (DECODE[2/3/4]BYTES_INDX_VALUETYPE), etc.
#include <anna/diameter/functions.hpp>
#include <anna/diameter/codec/functions.hpp> // REQUIRED_WORDS
#include <anna/diameter/codec/OamModule.hpp>
#include <anna/diameter/codec/Engine.hpp>
#include <anna/diameter/codec/EngineManager.hpp>
#include <anna/diameter/stack/Avp.hpp>
#include <anna/diameter/stack/Format.hpp>
#include <anna/diameter/stack/Dictionary.hpp>
#include <anna/diameter/stack/Engine.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/util/RegularExpression.hpp>
#include <anna/diameter/helpers/base/defines.hpp>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/xml/xml.hpp>

// STL
#include <string>
#include <vector>


using namespace anna;
using namespace anna::diameter::codec;


//static
namespace anna {

namespace diameter {

namespace codec {
const int Message::HeaderLength(20);
const U8 Message::RBitMask(0x80);
const U8 Message::PBitMask(0x40);
const U8 Message::EBitMask(0x20);
const U8 Message::TBitMask(0x10);
}
}
}

//------------------------------------------------------------------------------
//----------------------------------------------------------- Message::Message()
//------------------------------------------------------------------------------
Message::Message(Engine *engine) : a_engine(engine), a_forCode(true) {
  initialize();
}


//------------------------------------------------------------------------------
//----------------------------------------------------------- Message::Message()
//------------------------------------------------------------------------------
Message::Message(CommandId id, Engine *engine) : a_engine(engine), a_forCode(true) {
  initialize();
  setId(id);
}


//------------------------------------------------------------------------------
//---------------------------------------------------------- Message::~Message()
//------------------------------------------------------------------------------
Message::~Message() {
  clear();
}


//------------------------------------------------------------------------------
//--------------------------------------------------------- Message::setEngine()
//------------------------------------------------------------------------------
void Message::setEngine(Engine *engine) throw() {

  if (!engine) {
    LOGWARNING(anna::Logger::warning("Ignored: you must assign a valid codec engine. If you want to set NULL engine, clear the message", ANNA_FILE_LOCATION));
    return;
  }

  if (a_engine && (engine != a_engine)) {
    LOGWARNING(anna::Logger::warning("Ignored: it is not a good practice to change the codec engine once assigned. Clear the message first to set the engine again.", ANNA_FILE_LOCATION));
    return;
  }

  a_engine = engine;
}


//------------------------------------------------------------------------------
//--------------------------------------------------------- Message::getEngine()
//------------------------------------------------------------------------------
Engine * Message::getEngine() const throw(anna::RuntimeException) {
  if(!a_engine)
    throw anna::RuntimeException("Invalid codec engine reference (NULL). Use setEngine() to set the corresponding codec engine", ANNA_FILE_LOCATION);

  return a_engine;

}


//------------------------------------------------------------------------------
//-------------------------------------------------------- Message::initialize()
//------------------------------------------------------------------------------
void Message::initialize() throw() {
  a_version = 1;
  a_id = CommandId(0, false);
  a_flags = 0x00;
  a_applicationId = 0;
  a_hopByHop = 0;
  a_endToEnd = 0;
  //a_avps.clear();
  a_insertionPositionForChilds = 0;
  //a_finds.clear();
}


//------------------------------------------------------------------------------
//------------------------------------------------------------- Message::clear()
//------------------------------------------------------------------------------
void Message::clear(bool resetEngine) throw(anna::RuntimeException) {
  for(avp_iterator it = avp_begin(); it != avp_end(); it++) { /*avp(it)->clear(); */getEngine()->releaseAvp(Avp::avp(it)); }

  a_avps.clear();
  a_forCode.clear();
  // Cache system:
  a_finds.clear();
  // Initialize:
  initialize();
  if (resetEngine) a_engine = NULL;
}


//------------------------------------------------------------------------------
//----------------------------------------------------------- Message::flagsOK()
//------------------------------------------------------------------------------
bool Message::flagsOK(int &rc) const throw() {
  // Dictionary stack command:
  const stack::Command *stackCommand = getStackCommand();

  if(!stackCommand) {
    std::string msg = "Impossible to decide if flags are correct because stack command is not identified. Assume flags ok for Message ";
    msg += anna::diameter::functions::commandIdAsPairString(a_id);
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
    //rc = helpers::base::AVPVALUES__Result_Code::?????;
    return true;
  };

  // DIAMETER_INVALID_HDR_BITS 3008
  //
  //      A request was received whose bits in the Diameter header were set
  //      either to an invalid combination or to a value that is
  //      inconsistent with the Command Code's definition.
  bool ok = true;

  if(stackCommand->isRequest() != isRequest()) ok = false;  // en teoria es imposible salir por aqui: blindado en la dtd

  if(isRequest() && errorBit()) {
    std::string msg = "E(rror) bit is not allowed at diameter requests as ";
    msg += stackCommand->getName();
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
    ok = false;
  }

  if(isAnswer() && potentiallyReTransmittedMessageBit()) {
    std::string msg = "T(Potentially re-transmitted message) bit is not allowed at diameter answers as ";
    msg += stackCommand->getName();
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
    ok = false;
  }

  if(!ok) {
    rc = helpers::base::AVPVALUES__Result_Code::DIAMETER_INVALID_HDR_BITS;
    return false;
  }

  // DIAMETER_INVALID_BIT_IN_HEADER 5013
  //
  //      This error is returned when a reserved bit in the Diameter header
  //      is set to one (1) or the bits in the Diameter header are set
  //      incorrectly.
  if((a_flags & 0x0f) != 0x00) {
    std::string msg = "Any (or more than one) of the reserved message flags bit has been activated. Reserved bits must be null. Message is ";
    msg += stackCommand->getName();
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
    rc = helpers::base::AVPVALUES__Result_Code::DIAMETER_INVALID_BIT_IN_HEADER;
    return false;
  }

  return true;
}


//------------------------------------------------------------------------------
//------------------------------------------------------------- Message::setId()
//------------------------------------------------------------------------------
void Message::setId(CommandId id) throw(anna::RuntimeException) {

  // Id assignment:
  a_id = id;
  // Dictionary stack command:
  const stack::Command *stackCommand = getStackCommand(); // based on dictionary and a_id

  // Dictionary flags for known types:
  if(stackCommand) {
    if(stackCommand->isRequest()) a_flags |= RBitMask;
  } else {
    if(a_id.second) a_flags |= RBitMask; else a_flags &= (~RBitMask);
  }
}


//------------------------------------------------------------------------------
//------------------------------------------------------------- Message::setId()
//------------------------------------------------------------------------------
void Message::setId(const char *name) throw(anna::RuntimeException) {
  setId(getEngine()->commandIdForName(name));
}


//------------------------------------------------------------------------------
//-------------------------------------------------- Message::setApplicationId()
//------------------------------------------------------------------------------
void Message::setApplicationId(U32 aid) throw(anna::RuntimeException) {
  a_applicationId = aid;

  // Automatic engine configuration:
  if (a_engine) return;

  // Codec engine manager (a multithreaded application, normally does not achieve this point, because
  // messages are prepared for each interface with the corresponding codec engine)
  anna::diameter::codec::EngineManager &em = anna::diameter::codec::EngineManager::instantiate();
  if (em.size() == 0) return;
  if (em.selectFromApplicationId()) {
    Engine *monostackEngine = em.getMonoStackCodecEngine();
    if (monostackEngine) { a_engine = monostackEngine; return; }
    a_engine = em.getCodecEngine(aid);
  }
}


//------------------------------------------------------------------------------
//------------------------------------------------------------ Message::addAvp()
//------------------------------------------------------------------------------
Avp * Message::addAvp(const char *name) throw(anna::RuntimeException) {
  return addAvp(getEngine()->avpIdForName(name));
}


//------------------------------------------------------------------------------
//------------------------------------------------------------ Message::addAvp()
//------------------------------------------------------------------------------
Avp * Message::addAvp(Avp * avp) throw() {
  if(!avp) return NULL;
  if (avp->getEngine() != getEngine()) return NULL;
  addChild(avp);
  return avp;
}


//------------------------------------------------------------------------------
//--------------------------------------------------------- Message::removeAvp()
//------------------------------------------------------------------------------
bool Message::removeAvp(const char *name, int ocurrence) throw(anna::RuntimeException) {
  return removeAvp(getEngine()->avpIdForName(name), ocurrence);
}


//------------------------------------------------------------------------------
//----------------------------------------------------------- Message::_getAvp()
//------------------------------------------------------------------------------
const Avp * Message::_getAvp(const char *name, int ocurrence, anna::Exception::Mode::_v emode) const throw(anna::RuntimeException) {
  return getAvp(getEngine()->avpIdForName(name), ocurrence, emode);
}


//------------------------------------------------------------------------------
//---------------------------------------------------------- Message::countAvp()
//------------------------------------------------------------------------------
int Message::countAvp(const char *name) const throw(anna::RuntimeException) {
  return countAvp(getEngine()->avpIdForName(name));
}


//------------------------------------------------------------------------------
//--------------------------------------------------------- Message::getLength()
//------------------------------------------------------------------------------
U24 Message::getLength() const throw() {
  U24 result;
  // Header length:
  result = HeaderLength;

  for(const_avp_iterator it = avp_begin(); it != avp_end(); it++) result += 4 * REQUIRED_WORDS(Avp::avp(it)->getLength());

  return result;
}


//------------------------------------------------------------------------------
//------------------------------------------------------------ Message::decode()
//------------------------------------------------------------------------------
void Message::decode(const anna::DataBlock &db, Message *ptrAnswer) throw(anna::RuntimeException) {
  // Trace
  LOGDEBUG(
      anna::xml::Node root("Message::decode");
  std::string trace = "DataBlock to decode:\n";
  trace += db.asString();
  anna::Logger::debug(trace, ANNA_FILE_LOCATION);
  );
  clear(false /* respect engine */);
  // EXCEPTION MANAGEMENT IN THIS METHOD
  // ===================================
  // DECODE PHASE
  // If an error ocurred, decoding will stop launching exception but we will catch it and go on with validation because perhaps
  //  the achieved message could be valid against all odds. Only fatal errors cause direct decoding exception (length problems
  //  at header).
  //
  // VALIDATION PHASE
  // Launch exception on first validation error (validateAll == false), or log warning reporting all validation errors when
  //  complete validation is desired (validateAll == true, engine default) launching a final exception like "the decoded message is invalid".
  // OAM
  OamModule &oamModule = OamModule::instantiate();

  if(db.getSize() < HeaderLength) {
    oamModule.activateAlarm(OamModule::Alarm::MessageDecode__NotEnoughBytesToCoverMessageHeaderLength);
    oamModule.count(OamModule::Counter::MessageDecode__NotEnoughBytesToCoverMessageHeaderLength);
    // DIAMETER_INVALID_MESSAGE_LENGTH; // no podr� construir un answer fiable, as� que no registro el result-code
    throw anna::RuntimeException("Not enough bytes to cover message header length (20 bytes)", ANNA_FILE_LOCATION);
  }

  const char * buffer = db.getData();

  //         0                   1                   2                   3
  //         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
  //        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //        |    Version    |                 Message Length                |
  //        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //        | command flags |                  Command-Code                 |
  //        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //        |                         Application-ID                        |
  //        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //        |                      Hop-by-Hop Identifier                    |
  //        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //        |                      End-to-End Identifier                    |
  //        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //        |  AVPs ...
  //        +-+-+-+-+-+-+-+-+-+-+-+-+-
  // Header fields:
  a_version = (S8)buffer[0];

  U24 length = DECODE3BYTES_INDX_VALUETYPE(buffer, 1, U24); // total length including header fields

  a_flags = (S8)buffer[4];

  U24 code = DECODE3BYTES_INDX_VALUETYPE(buffer, 5, U24);

  a_id = CommandId(code, requestBit() /* based on a_flags */);

  setApplicationId(DECODE4BYTES_INDX_VALUETYPE(buffer, 8, U32)); // centralize set, because it could be used for stack selection.

  a_hopByHop = DECODE4BYTES_INDX_VALUETYPE(buffer, 12, U32);

  a_endToEnd = DECODE4BYTES_INDX_VALUETYPE(buffer, 16, U32);

  // Only build answer for a request:
  Message *answer = isRequest() ? ptrAnswer : NULL;

  if(answer) answer->setHeaderToAnswer(*this);

  // Length check:
  if(db.getSize() < length) {
    oamModule.activateAlarm(OamModule::Alarm::MessageDecode__NotEnoughBytesToCoverMessageLength);
    oamModule.count(OamModule::Counter::MessageDecode__NotEnoughBytesToCoverMessageLength);

    if(answer) answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_INVALID_MESSAGE_LENGTH);

    throw anna::RuntimeException("Not enough bytes to cover message length", ANNA_FILE_LOCATION);
  }

  // Message identifier
  // Flags could have been updated regarding dictionary, but during decoding we must respect buffer received:
  a_flags = (S8)buffer[4];
  // Avps start position:
  const S8 * startData = buffer + HeaderLength;
  U24 dataBytes = length - HeaderLength;

  if(dataBytes < 8 /* minimum avp */) {
    LOGDEBUG(anna::Logger::debug("Message empty (without avps)", ANNA_FILE_LOCATION));
    return;
  }

  int avpPos = 0;
  Avp* avp;
  anna::DataBlock db_aux;

  // Parent information:
  parent_t parent;
  parent.setMessage(a_id);


  while(avpPos < dataBytes) {
    try {
      avp =  getEngine()->createAvp(NULL);
      db_aux.assign(startData + avpPos, dataBytes - avpPos /* is valid to pass total length (indeed i don't know the real avp length) because it will be limited and this has deep copy disabled (no memory is reserved) */);
      avp -> decode(db_aux, parent, answer);
    } catch(anna::RuntimeException &ex) {
      getEngine()->releaseAvp(avp);
      LOGWARNING(
          anna::Logger::warning(ex.getText(), ANNA_FILE_LOCATION);
      anna::Logger::warning("Although a decoding error was found, validation could be checked because message could be enough for the application", ANNA_FILE_LOCATION);
      );
      break;
    }

    addChild(avp);
    avpPos += 4 * REQUIRED_WORDS(avp->getLength());
  }

  // Post-Fixing
  Engine::FixMode::_v fmode = getEngine()->getFixMode();

  if((fmode == Engine::FixMode::AfterDecoding) || (fmode == Engine::FixMode::Always)) fix();

  // Trace
  LOGDEBUG(
      std::string trace = "Message decoded:\n";
  trace += asXMLString();
  anna::Logger::debug(trace, ANNA_FILE_LOCATION);
  );
  // Post-Validation
  Engine::ValidationMode::_v vmode = getEngine()->getValidationMode();

  if((vmode == Engine::ValidationMode::AfterDecoding) || (vmode == Engine::ValidationMode::Always))
    if(!valid(answer))
      throw anna::RuntimeException("The decoded message is invalid. See previous report on warning-level traces", ANNA_FILE_LOCATION);
}


//------------------------------------------------------------------------------
//--------------------------------------------------- Message::getStackCommand()
//------------------------------------------------------------------------------
const anna::diameter::stack::Command *Message::getStackCommand(CommandId id) const throw(anna::RuntimeException) {
  const stack::Dictionary * dictionary = getEngine()->getDictionary();
  return (dictionary ? (dictionary->getCommand(id)) : NULL);
}


//------------------------------------------------------------------------------
//----------------------------------------------------- Message::setResultCode()
//------------------------------------------------------------------------------
void Message::setResultCode(int rc) throw(anna::RuntimeException) {
  if(isRequest()) return;

  // Add Result-Code if not yet added. Even if validation depth is set to 'Complete',
  // the Result-Code value will be the first found during the message analysis:
  Avp *resultCodeAvp = getAvp(helpers::base::AVPID__Result_Code, 1, anna::Exception::Mode::Ignore);

  if(!resultCodeAvp)
    addAvp(helpers::base::AVPID__Result_Code)->getUnsigned32()->setValue(rc);
  else if(resultCodeAvp->getUnsigned32()->getValue() == helpers::base::AVPVALUES__Result_Code::DIAMETER_SUCCESS)  // Only success could be replaced
    resultCodeAvp->getUnsigned32()->setValue(rc);

  // Error bit:
  setErrorBit(rc >= 3001 && rc <= 3010 /* protocol errors */);
}


//------------------------------------------------------------------------------
//----------------------------------------------------- Message::getResultCode()
//------------------------------------------------------------------------------
int Message::getResultCode() const throw() {
  if(isAnswer()) {
    const Avp *resultCodeAvp = getAvp(helpers::base::AVPID__Result_Code, 1, anna::Exception::Mode::Ignore);

    if(resultCodeAvp)
      return resultCodeAvp->getUnsigned32()->getValue();
  }

  return -1;
}


//------------------------------------------------------------------------------
//------------------------------------------------------ Message::setFailedAvp()
//------------------------------------------------------------------------------
void Message::setFailedAvp(const parent_t &parent, AvpId wrong, const char *wrongName) throw(anna::RuntimeException) {

  if(isRequest()) return;

  // RFC 6733:
  //
  //  7.5.  Failed-AVP AVP
  //
  //     The Failed-AVP AVP (AVP Code 279) is of type Grouped and provides
  //     debugging information in cases where a request is rejected or not
  //     fully processed due to erroneous information in a specific AVP.  The
  //     value of the Result-Code AVP will provide information on the reason
  //     for the Failed-AVP AVP.  A Diameter answer message SHOULD contain an
  //     instance of the Failed-AVP AVP that corresponds to the error
  //     indicated by the Result-Code AVP.  For practical purposes, this
  //     Failed-AVP would typically refer to the first AVP processing error
  //     that a Diameter node encounters.

  // Although the Failed-AVP definition has cardinality 1* and Failed-AVP itself is defined in
  // most of the command codes as *[Failed-AVP], i think this is not a deliberate ambiguity.
  // Probably the RFC wants to give freedom to the application layer, but it is recommended to
  // have only one child (wrong avp) inside a unique message Failed-AVP to ease the Result-Code
  // correspondence. Anyway, this behaviour could be easily  opened by mean 'setSingleFailedAVP(false)'
  Avp *theFailedAvp = getAvp(helpers::base::AVPID__Failed_AVP, 1, anna::Exception::Mode::Ignore);
  if (theFailedAvp) {
    if (getEngine()->getSingleFailedAVP()) {
      LOGDEBUG(anna::Logger::debug("Failed-AVP has already been added. RFC 6733 Section 7.5 recommends to store only the first error found", ANNA_FILE_LOCATION));
      return;
    }
  }

  // Section 7.5 RFC 6733: A Diameter message SHOULD contain one Failed-AVP AVP
  theFailedAvp = addAvp(helpers::base::AVPID__Failed_AVP);
  Avp *leaf = theFailedAvp;

  LOGDEBUG(
      std::string msg = "Adding to Failed-AVP, the wrong avp ";
  msg += wrongName ? wrongName : (anna::diameter::functions::avpIdAsPairString(wrong));
  msg += " found inside ";
  msg += parent.asString();

  anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );

  std::vector<AvpId>::const_iterator it;
  for(it = parent.AvpsId.begin(); it != parent.AvpsId.end(); it++)
    leaf = leaf->addAvp(*it);

  leaf->addAvp(wrong);
}


//------------------------------------------------------------------------------
//----------------------------------------------- Message::setStandardToAnswer()
//------------------------------------------------------------------------------
void Message::setStandardToAnswer(const Message &request, const std::string &originHost, const std::string &originRealm, int resultCode) throw(anna::RuntimeException) {
  if(!request.getId().second) return;

  // Message header:
  setHeaderToAnswer(request);
  // Session-Id if exists:
  const Avp *reqSessionId = request.getAvp(helpers::base::AVPID__Session_Id, 1, anna::Exception::Mode::Ignore);

  LOGDEBUG(anna::Logger::debug("Check answer message AVPs Session-Id, Origin-Host and Origin-Realm => replace them if missing, with request session-id & node configuration:", ANNA_FILE_LOCATION));

  if(reqSessionId)
    if(!getAvp(helpers::base::AVPID__Session_Id, 1, anna::Exception::Mode::Ignore))
      addAvp(helpers::base::AVPID__Session_Id)->getUTF8String()->setValue(reqSessionId->getUTF8String()->getValue());

  // Origin-Host & Realm
  if(!getAvp(helpers::base::AVPID__Origin_Host, 1, anna::Exception::Mode::Ignore))
    addAvp(helpers::base::AVPID__Origin_Host)->getDiameterIdentity()->setValue(originHost);

  if(!getAvp(helpers::base::AVPID__Origin_Realm, 1, anna::Exception::Mode::Ignore))
    addAvp(helpers::base::AVPID__Origin_Realm)->getDiameterIdentity()->setValue(originRealm);

  // Proxy-Info AVPs if exist, in the same order:
  for(const_avp_iterator it = request.avp_begin(); it != request.avp_end(); it++)
    if((*it).second->getId() == helpers::base::AVPID__Proxy_Info)
      addAvp((*it).second);

  // If no Result-Code was added is because all was ok, then application could detect another problem:
  setResultCode(resultCode);
  // Fix:
  fix();
  LOGDEBUG(
      std::string msg = "Completed answer:\n";
  msg += asXMLString();
  anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}


//------------------------------------------------------------------------------
//--------------------------------------------------------------- Message::fix()
//------------------------------------------------------------------------------
void Message::fix() throw() {
  // Dictionary stack command:
  const stack::Command *stackCommand = getStackCommand();

  if(!stackCommand) {
    LOGDEBUG(anna::Logger::debug("No dictionary command reference found. Cannot fix.", ANNA_FILE_LOCATION));
    // Try to get to the first place the Session-ID (if exists) because...
    // RFC 6733: All messages pertaining to a specific session MUST include only one Session-Id AVP ...
    //           ... When present, the Session-Id SHOULD appear immediately following the Diameter header
    avp_iterator it = Avp::avp_find(a_avps, helpers::base::AVPID__Session_Id, 1 /* one and only */);

    if(it != avp_end()) {
      int sessionPos = (*it).first;
      Avp *first = (*avp_begin()).second;
      a_avps[0] = (*it).second; // Session-Id
      a_avps[sessionPos] = first;
      LOGDEBUG(anna::Logger::debug("Session-Id has been manually fixed to the first place", ANNA_FILE_LOCATION));
    }

    return;
  }

  Avp::fix(a_avps, (find_container&)a_finds, a_insertionPositionForChilds, stackCommand->avprule_begin(), stackCommand->avprule_end());
}


//------------------------------------------------------------------------------
//------------------------------------------------------------- Message::valid()
//------------------------------------------------------------------------------
bool Message::valid(Message *ptrAnswer) const throw(anna::RuntimeException) {
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  // Dictionary stack command:
  const stack::Command *stackCommand = getStackCommand();
  // Only build answer for a request:
  Message *answer = isRequest() ? ptrAnswer : NULL;

  if(answer) answer->setHeaderToAnswer(*this);

  if(!stackCommand) {
    // OAM
    std::string me = anna::diameter::functions::commandIdAsPairString(a_id);
    oamModule.activateAlarm(OamModule::Alarm::MessageValidation__UnknownOperation__s__UnableToValidate, STRING_WITH_QUOTATION_MARKS__C_STR(me));
    oamModule.count(OamModule::Counter::MessageValidation__UnknownOperationUnableToValidate);

    if(answer) answer->setResultCode(helpers::base::AVPVALUES__Result_Code::DIAMETER_COMMAND_UNSUPPORTED);

    getEngine()->validationAnomaly(anna::functions::asString("Unknown operation %s. Unable to validate", STRING_WITH_QUOTATION_MARKS__C_STR(me)));
    return false;
  }

  // Parent information:
  parent_t me;
  me.setMessage(a_id, stackCommand->getName().c_str());

  //////////////////////////////
  // Flags coherence checking //
  //////////////////////////////
  int rc;
  bool result = flagsOK(rc);

  if(!result) {
    // OAM & Depth management
    oamModule.activateAlarm(OamModule::Alarm::MessageValidation__Operation__s__HaveIncoherentFlags__d__, STRING_WITH_QUOTATION_MARKS__C_STR(me.asString()), (int)a_flags);
    oamModule.count(OamModule::Counter::MessageValidation__OperationHaveIncoherentFlags);

    if(answer) answer->setResultCode(rc);

    getEngine()->validationAnomaly(anna::functions::asString("Operation %s have incoherent flags (%d)", STRING_WITH_QUOTATION_MARKS__C_STR(me.asString()), (int)a_flags));
  }

  ////////////////////
  // Level checking //
  ////////////////////
  result = Avp::validLevel(a_avps, stackCommand->avprule_begin(), stackCommand->avprule_end(), getEngine(), me, answer) && result;

  ////////////////////////
  // Childrens checking //
  ////////////////////////
  for(const_avp_iterator it = avp_begin(); it != avp_end(); it++)
    result = ((*it).second->valid(me, answer)) && result;

  return result;
}


//------------------------------------------------------------------------------
//-------------------------------------------------------------- Message::code()
//------------------------------------------------------------------------------
const anna::DataBlock & Message::code() throw(anna::RuntimeException) {
  // Pre-Validation
  Engine::ValidationMode::_v vmode = getEngine()->getValidationMode();

  if((vmode == Engine::ValidationMode::BeforeEncoding) || (vmode == Engine::ValidationMode::Always)) {
    if(!valid())
      throw anna::RuntimeException("Try to encode an invalid message. See previous report on warning-level traces", ANNA_FILE_LOCATION);
  }

  // Pre-Fixing
  Engine::FixMode::_v fmode = getEngine()->getFixMode();

  if((fmode == Engine::FixMode::BeforeEncoding) || (fmode == Engine::FixMode::Always)) fix();

  // Trace
  LOGDEBUG(
      std::string trace = "Message to code:\n";
  trace += asXMLString();
  anna::Logger::debug(trace, ANNA_FILE_LOCATION);
  );
  // Memory allocation
  U24 length = getLength();
  a_forCode.clear();
  a_forCode.allocate(length);
  char* buffer = (char*)a_forCode.getData();
  // Version and length
  buffer[0] = (S8) a_version;
  buffer[1] = (S8)(length >> 16);
  buffer[2] = (S8)(length >> 8);
  buffer[3] = (S8) length;
  // Flags and code
  buffer[4] = (S8) a_flags;
  buffer[5] = (S8)(a_id.first >> 16);
  buffer[6] = (S8)(a_id.first >> 8);
  buffer[7] = (S8) a_id.first;
  // Application Id
  buffer[8] = (S8)(a_applicationId >> 24);
  buffer[9] = (S8)(a_applicationId >> 16);
  buffer[10] = (S8)(a_applicationId >> 8);
  buffer[11] = (S8) a_applicationId;
  // Hob by hop
  buffer[12] = (S8)(a_hopByHop >> 24);
  buffer[13] = (S8)(a_hopByHop >> 16);
  buffer[14] = (S8)(a_hopByHop >> 8);
  buffer[15] = (S8) a_hopByHop;
  // End to end
  buffer[16] = (S8)(a_endToEnd >> 24);
  buffer[17] = (S8)(a_endToEnd >> 16);
  buffer[18] = (S8)(a_endToEnd >> 8);
  buffer[19] = (S8) a_endToEnd;
  // Data start position:
  int startDataPos = HeaderLength;

  if(startDataPos == length) {
    LOGDEBUG(anna::Logger::debug("There is no Avps to encode (only-header message)", ANNA_FILE_LOCATION));
  } else {
    // Data part:
    S8 * dataPart = buffer + startDataPos;
    int dataBytes; // not used but could be useful for checking (length - startDataPos)
    // Each avp encoding will remain padding octets depending on format. In order to
    //  code avps colection (each avp will be multiple of 4) we clean the entire buffer
    //  to ensure padding (easier than custom-made for each format):
    memset(dataPart, 0, length - startDataPos); // no estoy seguro de que el clear del DataBlock haga esto...

    for(const_avp_iterator it = avp_begin(); it != avp_end(); it++) {
      Avp::avp(it)->code(dataPart, dataBytes);
      dataPart = dataPart + 4 * REQUIRED_WORDS(dataBytes);
    }
  }

  // Trace
  LOGDEBUG(
      std::string trace = "DataBlock encoded:\n";
  trace += a_forCode.asString();
  //      trace += "\nAs continuous hexadecimal string:\n";
  //      trace += anna::functions::asHexString(a_forCode);
  anna::Logger::debug(trace, ANNA_FILE_LOCATION);
  );
  return a_forCode;
}

//------------------------------------------------------------------------------
//------------------------------------------------------- Message::loadXMLFile()
//------------------------------------------------------------------------------
void Message::loadXMLFile(const std::string &xmlPathFile) throw(anna::RuntimeException) {

  anna::xml::DocumentFile xmlDocument;
  anna::diameter::codec::functions::messageXmlDocumentFromXmlFile(xmlDocument, xmlPathFile);
  fromXML(xmlDocument.getRootNode());
}

//------------------------------------------------------------------------------
//----------------------------------------------------- Message::loadXMLString()
//------------------------------------------------------------------------------
void Message::loadXMLString(const std::string &xmlString) throw(anna::RuntimeException) {

  anna::xml::DocumentMemory xmlDocument;
  anna::diameter::codec::functions::messageXmlDocumentFromXmlString(xmlDocument, xmlString);
  fromXML(xmlDocument.getRootNode());
}

//------------------------------------------------------------------------------
//----------------------------------------------------------- Message::fromXML()
//------------------------------------------------------------------------------
void Message::fromXML(const anna::xml::Node* messageNode) throw(anna::RuntimeException) {
  // <!ATTLIST message version CDATA #IMPLIED name CDATA #IMPLIED code CDATA #IMPLIED flags CDATA #IMPLIED p-bit (yes | no) #IMPLIED e-bit (yes | no) #IMPLIED t-bit (yes | no) #IMPLIED application-id CDATA #REQUIRED hop-by-hop-id CDATA #IMPLIED end-to-end-id CDATA #IMPLIED>
  const anna::xml::Attribute *version, *name, *code, *flags, *pbit, *ebit, *tbit, *appid, *hbh, *ete;
  version = messageNode->getAttribute("version", false /* no exception */);
  name = messageNode->getAttribute("name", false /* no exception */);
  code = messageNode->getAttribute("code", false /* no exception */);
  flags = messageNode->getAttribute("flags", false /* no exception */);
  pbit = messageNode->getAttribute("p-bit", false /* no exception */);
  ebit = messageNode->getAttribute("e-bit", false /* no exception */);
  tbit = messageNode->getAttribute("t-bit", false /* no exception */);
  appid = messageNode->getAttribute("application-id"); // required
  hbh = messageNode->getAttribute("hop-by-hop-id", false /* no exception */);
  ete = messageNode->getAttribute("end-to-end-id", false /* no exception */);
  int i_aux;
  unsigned int u_aux;

  // Clear the message
  clear(false /* respect engine */);

  if(version) {
    i_aux = version->getIntegerValue();

    if(i_aux < 0 || i_aux > 256) {
      std::string msg = "Error processing avp <version '"; msg += version->getValue();
      msg += "': out of range [0,256]";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    a_version = i_aux;
  }

  // Application-id
  // This is called before any operation which needs to know about the stack elements (this could set the dictionary)
  setApplicationId(appid->getIntegerValue());

  // Dictionary
  const stack::Dictionary * dictionary = getEngine()->getDictionary();
  const stack::Command *stackCommand = NULL;

  // Compact mode
  if(name) {
    if(!dictionary) {
      std::string msg = "Error processing command <name '"; msg += name->getValue();
      msg += "'>: no dictionary available. Load one or use <'code' + 'flags'> instead of <'name'>";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    stackCommand = dictionary->getCommand(name->getValue());

    if(!stackCommand) {
      std::string msg = "Error processing command <name '"; msg += name->getValue();
      msg += "'>: no command found for this identifier at available dictionary";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }
  }

  // Check attributes exclusiveness
  if(name) {  // compact mode
    if(code || flags) {
      std::string msg = "Error processing command <name '"; msg += name->getValue();
      msg += "'>: message attributes <'code' + 'flags'> are not allowed if <'name'> is provided";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    setId(stackCommand->getId());

    // 'P', 'E' and 'T' flags:
    bool activateP = pbit ? (pbit->getValue() == "yes") : false;
    bool activateE = ebit ? (ebit->getValue() == "yes") : false;
    bool activateT = tbit ? (tbit->getValue() == "yes") : false;

    if(activateP) a_flags |= PBitMask;

    if(activateE) a_flags |= EBitMask;

    if(activateT) a_flags |= TBitMask;
  } else {
    std::string s_code = code ? code->getValue() : "?";
    std::string s_flags = flags ? flags->getValue() : "?";

    if(!code || !flags) {
      std::string msg = "Error processing command <code '"; msg += s_code;
      msg += "' + flags '"; msg += s_flags;
      msg += "'>: both must be provided if <'name'> don't";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    if(pbit || ebit || tbit) {
      std::string msg = "Error processing command <code '"; msg += s_code;
      msg += "' + flags '"; msg += s_flags;
      msg += "'>: neither of 'p-bit', 'e-bit' or 't-bit' fields are allowed if those are present";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    // Code check
    i_aux = code->getIntegerValue();

    if(i_aux < 0) {
      std::string msg = "Error processing command <code '"; msg += s_code;
      msg += "': negative values are not allowed";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    U24 u_code = i_aux;
    // Flags check
    i_aux = flags->getIntegerValue();

    if(i_aux < 0 || i_aux > 256) {
      std::string msg = "Error processing command <flags '"; msg += s_flags;
      msg += "': out of range [0,256]";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    a_flags = i_aux;
    int flagsBCK = a_flags;
    // Final assignments
    a_id = CommandId(u_code, requestBit() /* based on a_flags */);
    // Flags could have been updated regarding dictionary, but during parsing we must respect xml file:
    a_flags = flagsBCK;
  }

  // Hob-by-hop-id
  if(hbh) {
    u_aux = hbh->getIntegerValue();

    /*
    if(u_aux < 0) {
      std::string msg = "Error processing command <hop-by-hop-id '"; msg += hbh->getValue();
      msg += "': negative values are not allowed";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }
     */
  } else u_aux = 0;

  setHopByHop(u_aux);

  // End-to-end-id
  if(ete) {
    u_aux = ete->getIntegerValue();

    /*
    if(u_aux < 0) {
      std::string msg = "Error processing command <end-to-end-id '"; msg += ete->getValue();
      msg += "': negative values are not allowed";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }
     */
  } else u_aux = 0;

  setEndToEnd(u_aux);
  // Childrens
  Avp *avp;

  for(anna::xml::Node::const_child_iterator it = messageNode->child_begin(); it != messageNode->child_end(); it++) {
    std::string nodeName = (*it)->getName();

    if(nodeName != "avp") {
      std::string msg = "Unsupported message child node name '"; msg += nodeName;
      msg += "': use 'avp'";
      throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    try {
      avp =  getEngine()->createAvp(NULL);
      avp -> fromXML(*it);
    } catch(anna::RuntimeException &ex) {
      getEngine()->releaseAvp(avp);
      throw ex;
    }

    addAvp(avp);
  }
}


//------------------------------------------------------------------------------
//------------------------------------------------------------- Message::asXML()
//------------------------------------------------------------------------------
anna::xml::Node* Message::asXML(anna::xml::Node* parent) const throw() {
  // <!ATTLIST message version CDATA #IMPLIED name CDATA #IMPLIED code CDATA #IMPLIED flags CDATA #IMPLIED application-id CDATA #REQUIRED hop-by-hop-id CDATA #IMPLIED end-to-end-id CDATA #IMPLIED>
  anna::xml::Node* result = parent->createChild("message");
  // Dictionary stack command:
  const stack::Command *stackCommand = getStackCommand();
  bool compactMode = stackCommand /*&& flagsOK()*/;
  result->createAttribute("version", anna::functions::asString((int)a_version));

  if(compactMode) {
    result->createAttribute("name", stackCommand->getName());

    if(proxiableBit()) result->createAttribute("p-bit", "yes");

    if(errorBit()) result->createAttribute("e-bit", "yes");

    if(potentiallyReTransmittedMessageBit()) result->createAttribute("t-bit", "yes");
  } else {
    result->createAttribute("code", a_id.first);
    result->createAttribute("flags", (int)a_flags);
  }

  result->createAttribute("application-id", anna::functions::asString(a_applicationId));
  result->createAttribute("hop-by-hop-id", anna::functions::asString(a_hopByHop));
  result->createAttribute("end-to-end-id", anna::functions::asString(a_endToEnd));

  // Avps:
  for(const_avp_iterator it = avp_begin(); it != avp_end(); it++) {
    Avp::avp(it)->asXML(result);
  }

  return result;
}


//------------------------------------------------------------------------------
//------------------------------------------------------- Message::asXMLString()
//------------------------------------------------------------------------------
std::string Message::asXMLString() const throw() {
  anna::xml::Node root("root");
  return anna::xml::Compiler().apply(asXML(&root));
}


//------------------------------------------------------------------------------
//------------------------------------------------------------ Message::isLike()
//------------------------------------------------------------------------------
bool Message::isLike(const std::string &pattern) const throw() {
  anna::RegularExpression re(pattern);
  return re.isLike(asXMLString());
}
